// JPS.hpp
//
//     Created on: 19. 02. 2018
//         Author: Jan Uhlik
//    Modified by:
//
// Copyright (c) 2017 Czech Technical University in Prague | Faculty of Information Technology. All rights reserved.
// Git repository: https://gitlab.fit.cvut.cz/algorithms-library-toolkit/automata-library

#ifndef ALIB2_JPS_HPP
#define ALIB2_JPS_HPP

#include <alib/vector>
#include <alib/pair>
#include <alib/set>
#include <alib/map>
#include <functional>
#include <limits>
#include <algorithm>

#include <common/ReconstructPath.hpp>
#include <common/SupportFunction.hpp>
#include <grid/common/GridFunction.hpp>

namespace graph {

namespace shortest_path {

class JPS {
// ---------------------------------------------------------------------------------------------------------------------

 public:
  /// Find the shortest path using Jump point search algorithm from the \p start node to the \p goal node in the \p graph.
  ///
  /// Whenever node is opened, \p f_user is called with two parameters (the opened node and value of currently shortest path).
  ///
  /// The heuristic function must be admissible and monotone.
  ///
  /// \param graph to explore.
  /// \param start initial node.
  /// \param goal final node.
  /// \param f_heuristic heuristic function which accept node and return edge_type::weight_type
  /// \param f_user function which is called for every opened node with value of currently shortest path.
  ///
  /// \returns pair where first := shortest path := distance of path, if there is no such path vector is empty and distance std::numeric_limits<edge_type:weight_type>::max()
  ///
  /// \node TGraph must follow graph::type::grid::SquareGrid8Weighted
  /// \sa graph::type::grid::SquareGri8Weighted
  /// \sa graph::type::grid::SquareGridInterface
  ///
  /// \note TEdge of \p graph must follow graph::edge::WeightedEdge interface
  /// \sa graph::edge_type::WeightedEdge
  ///
  ///
  template<
      typename TGraph,
      typename TNode,
      typename F1 = std::function<typename TGraph::edge_type::weight_type(const TNode &)>,
      typename F2 = std::function<void(const TNode &, const typename TGraph::edge_type::weight_type &)>
  >
  static
  ext::pair<ext::vector<TNode>, typename TGraph::edge_type::weight_type>
  findPath(const TGraph &graph,
           const TNode &start,
           const TNode &goal,
           F1 f_heuristic,
           F2 f_user = [](const TNode &,
                          const typename TGraph::edge_type::weight_type &) {});

  template<
      typename TGraph,
      typename TNode,
      typename F1 = std::function<typename TGraph::edge_type::weight_type(const TNode &, const TNode &)>
  >
  static
  ext::pair<ext::vector<TNode>, typename TGraph::edge_type::weight_type>
  findPathRegistration(const TGraph &graph,
                       const TNode &start,
                       const TNode &goal,
                       F1 f_heuristic) {
    return findPath(graph, start, goal, [&](const TNode &n) { return f_heuristic(goal, n); });
  }

// =====================================================================================================================

 private:

  template<typename TNode, typename TWeight>
  struct Data {
    ext::set<ext::pair<TWeight, TNode>> queue; // priority queue
    ext::map<TNode, TWeight> g; // G score
    ext::map<TNode, TNode> p; // parents
  };

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode, typename F1, typename F2, typename F3>
  static
  bool
  relaxation(const TGraph &graph,
             Data<TNode, typename TGraph::edge_type::weight_type> &data,
             const TNode &start,
             const TNode &goal,
             F1 f_heuristic,
             F2 f_user,
             F3 f_stop);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  static
  ext::map<TNode, typename TGraph::edge_type::weight_type>
  successors(const TGraph &graph,
             Data<TNode, typename TGraph::edge_type::weight_type> &data,
             const TNode &x,
             const TNode &start,
             const TNode &goal);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  static
  ext::map<TNode, typename TGraph::edge_type::weight_type>
  prune_neighbours(const TGraph &graph,
                   Data<TNode, typename TGraph::edge_type::weight_type> &data,
                   const TNode &n);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  static
  void
  jump(const TGraph &graph,
       ext::map<TNode, typename TGraph::edge_type::weight_type> &successors,
       const TNode &x,
       typename TGraph::direction_type direction,
       typename TGraph::edge_type::weight_type distance,
       const TNode &start,
       const TNode &goal);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TNode, typename TWeight, typename F>
  inline static
  void
  init(JPS::Data<TNode, TWeight> &data, const TNode &start, F f_heuristic);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  inline static
  void
  addNeighbour(const TGraph &graph,
               ext::map<TNode, typename TGraph::edge_type::weight_type> &neighbours,
               const TNode &n,
               typename TGraph::direction_type direction);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  inline static
  void
  forcedNeighbour(const TGraph &graph,
                  ext::map<TNode, typename TGraph::edge_type::weight_type> &neighbours,
                  const TNode &n,
                  typename TGraph::direction_type direction);

// ---------------------------------------------------------------------------------------------------------------------

  template<typename TGraph, typename TNode>
  static
  ext::vector<TNode>
  reconstructFullPath(const TGraph &graph,
                      Data<TNode, typename TGraph::edge_type::weight_type> &data,
                      const TNode &start,
                      const TNode &goal);


// ---------------------------------------------------------------------------------------------------------------------
};

// =====================================================================================================================

template<typename TGraph, typename TNode, typename F1, typename F2>
ext::pair<ext::vector<TNode>, typename TGraph::edge_type::weight_type>
JPS::findPath(const TGraph &graph,
              const TNode &start,
              const TNode &goal,
              F1 f_heuristic,
              F2 f_user) {

  using weight_type = typename TGraph::edge_type::weight_type;

  Data<TNode, weight_type> data;

  // Init search
  init(data, start, f_heuristic);

  while (!data.queue.empty()) {
    bool stop = relaxation(graph,
                           data,
                           start,
                           goal,
                           f_heuristic,
                           f_user,
                           [&](const TNode &node) -> bool { return goal == node; });

    if (stop) {
      break;
    }
  }

  if (data.g.find(goal) != data.g.end()) {
    return ext::make_pair(reconstructFullPath(graph, data, start, goal), data.g.at(goal));
  }

  return ext::make_pair(ext::vector<TNode>(), std::numeric_limits<weight_type>::max());
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode, typename F1, typename F2, typename F3>
bool
JPS::relaxation(const TGraph &graph,
                Data<TNode, typename TGraph::edge_type::weight_type> &data,
                const TNode &start,
                const TNode &goal,
                F1 f_heuristic,
                F2 f_user,
                F3 f_stop) {
  using weight_type = typename TGraph::edge_type::weight_type;

  TNode n = data.queue.begin()->second;
  data.queue.erase(data.queue.begin());

  // Run user's function
  f_user(n, data.g[n]);

  // Stop if reach the goal
  if (f_stop(n)) {
    return true;
  }

  ext::map<TNode, weight_type> succs = successors(graph, data, n, start, goal);
  for (const auto &s_pair: succs) {
    const TNode &s = s_pair.first; // successor
    const weight_type &distance = s_pair.second; // distance to successor

    // Calculate new G score
    weight_type gscore = data.g.at(n) + distance;

    // Search if the node s was already visited
    if (data.g.find(s) == data.g.end()) {
      data.g[s] = gscore;
      data.p.insert_or_assign(s, n);
      data.queue.insert(ext::make_pair(gscore + f_heuristic(s), s));
    }
  }

  return false;
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
ext::map<TNode, typename TGraph::edge_type::weight_type>
JPS::successors(const TGraph &graph,
                Data<TNode, typename TGraph::edge_type::weight_type> &data,
                const TNode &x,
                const TNode &start,
                const TNode &goal) {
  using weight_type = typename TGraph::edge_type::weight_type;

  ext::map<TNode, weight_type> successors;
  ext::map<TNode, weight_type> neighbours = prune_neighbours(graph, data, x);

  for (auto &n: neighbours) {
    jump(graph, successors, x, grid::GridFunction::squareGridDirection(x, n.first), n.second, start, goal);
  }

  return successors;
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
ext::map<TNode, typename TGraph::edge_type::weight_type>
JPS::prune_neighbours(const TGraph &graph,
                      JPS::Data<TNode, typename TGraph::edge_type::weight_type> &data,
                      const TNode &n) {
  using weight_type = typename TGraph::edge_type::weight_type;
  using direction_type = typename TGraph::direction_type;

  ext::map<TNode, weight_type> neighbours;

  direction_type direction = grid::GridFunction::squareGridDirection(data.p.at(n), n);
  if (grid::GridFunction::sqaureGridDirectionIsCardinal(direction)) {
    // Prune rules
    addNeighbour(graph, neighbours, n, direction);

    // Force neighbours
    forcedNeighbour(graph, neighbours, n, direction);

  } else if (grid::GridFunction::sqaureGridDirectionIsDiagonal(direction)) {
    // Prune rules
    addNeighbour(graph, neighbours, n, direction);
    for (const auto &i: grid::GridFunction::squareGridDirectionDecompose(direction)) {
      addNeighbour(graph, neighbours, n, i);
    }

    // Force neighbours
    forcedNeighbour(graph, neighbours, n, direction);

  } else { // In case that node n == start -> need to expand all neighbours
    for (const auto &s_edge : graph.successorEdges(n)) {
      const TNode &s = common::SupportFunction::other(s_edge, n); // successor
      neighbours[s] = s_edge.weight();
    }
  }

  return neighbours;
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
void JPS::jump(const TGraph &graph,
               ext::map<TNode, typename TGraph::edge_type::weight_type> &successors,
               const TNode &x,
               typename TGraph::direction_type direction,
               typename TGraph::edge_type::weight_type distance,
               const TNode &start,
               const TNode &goal) {
  using weight_type = typename TGraph::edge_type::weight_type;

  ext::pair<bool, TNode> step = graph.step(x, direction);
  bool &valid = step.first; // Nasty way because there is no support for C++17
  TNode &n = step.second; // Nasty way because there is no support for C++17

  if (!valid) {
    return;
  }

  if (n == goal) {
    successors[n] = distance;
    return;
  }

  ext::map<TNode, weight_type> forced_neighbours;
  forcedNeighbour(graph, forced_neighbours, n, direction);

  if (!forced_neighbours.empty ( )) {
    successors[n] = distance;
    return;
  }

  if (grid::GridFunction::sqaureGridDirectionIsDiagonal(direction)) {
    for (const auto &i: grid::GridFunction::squareGridDirectionDecompose(direction)) {
      ext::map<TNode, weight_type> tmp;
      jump(graph, tmp, n, i, 0, start, goal);

      if (!tmp.empty ( )) {
        successors[n] = distance;
        return;
      }
    }
  }

  return jump(graph,
              successors,
              n,
              direction,
              distance + graph.stepEdge(n, direction).second.weight(), // If not valid doesn't matter.
              start,
              goal);
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TNode, typename TWeight, typename F>
void JPS::init(JPS::Data<TNode, TWeight> &data,
               const TNode &start,
               F f_heuristic) {
  data.g[start] = 0;
  data.p.insert_or_assign(start, start);
  data.queue.insert(ext::make_pair(data.g[start] + f_heuristic(start), start));
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
void JPS::addNeighbour(const TGraph &graph,
                       ext::map<TNode, typename TGraph::edge_type::weight_type> &neighbours,
                       const TNode &n,
                       typename TGraph::direction_type direction) {
  using edge_type = typename TGraph::edge_type;

  ext::pair<bool, edge_type> successor_edge_pair = graph.stepEdge(n, direction);

  if (successor_edge_pair.first) {
    const TNode &s = common::SupportFunction::other(successor_edge_pair.second, n); // successor
    neighbours[s] = successor_edge_pair.second.weight();
  }
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
void JPS::forcedNeighbour(const TGraph &graph,
                          ext::map<TNode, typename TGraph::edge_type::weight_type> &neighbours,
                          const TNode &n,
                          typename TGraph::direction_type direction) {
  using direction_type = typename TGraph::direction_type;

  if (direction == direction_type::north) {
    if (!graph.step(n, direction_type::west).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_west);
    }
    if (!graph.step(n, direction_type::east).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_east);
    }
  } else if (direction == direction_type::east) {
    if (!graph.step(n, direction_type::north).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_east);
    }
    if (!graph.step(n, direction_type::south).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_east);
    }
  } else if (direction == direction_type::south) {
    if (!graph.step(n, direction_type::east).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_east);
    }
    if (!graph.step(n, direction_type::west).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_west);
    }
  } else if (direction == direction_type::west) {
    if (!graph.step(n, direction_type::north).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_west);
    }
    if (!graph.step(n, direction_type::south).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_west);
    }
  } else if (direction == direction_type::north_west) {
    if (!graph.step(n, direction_type::south).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_west);
    }
    if (!graph.step(n, direction_type::east).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_east);
    }
  } else if (direction == direction_type::north_east) {
    if (!graph.step(n, direction_type::south).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_east);
    }
    if (!graph.step(n, direction_type::west).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_west);
    }
  } else if (direction == direction_type::south_east) {
    if (!graph.step(n, direction_type::north).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_east);
    }
    if (!graph.step(n, direction_type::west).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_west);
    }
  } else if (direction == direction_type::south_west) {
    if (!graph.step(n, direction_type::north).first) {
      addNeighbour(graph, neighbours, n, direction_type::north_west);
    }
    if (!graph.step(n, direction_type::east).first) {
      addNeighbour(graph, neighbours, n, direction_type::south_east);
    }
  }
}

// ---------------------------------------------------------------------------------------------------------------------

template<typename TGraph, typename TNode>
ext::vector<TNode>
JPS::reconstructFullPath(const TGraph &graph,
                         JPS::Data<TNode, typename TGraph::edge_type::weight_type> &data,
                         const TNode &start,
                         const TNode &goal) {
  ext::vector<TNode> path;
  if (data.p.find(goal) == data.p.end()) {
    // Path not found -> return empty vector
    return path;
  }

  path.push_back(goal);
  TNode current_vertex = goal;

  while (current_vertex != start) {
    TNode jump_point = data.p.at(current_vertex);
    current_vertex =
        graph.step(current_vertex, grid::GridFunction::squareGridDirection(current_vertex, jump_point)).second;

    while (jump_point != current_vertex) {
      path.push_back(current_vertex);
      current_vertex =
          graph.step(current_vertex, grid::GridFunction::squareGridDirection(current_vertex, jump_point)).second;
    }

    path.push_back(jump_point);
  }

  // Path need to be reverse.
  std::reverse(path.begin(), path.end());

  return path;
}

// ---------------------------------------------------------------------------------------------------------------------

} // namespace shortest_path

} // namespace graph

#endif //ALIB2_JPS_HPP
